"""Nav Parrot Meta information."""

__title__ = "ai-parrot"
__description__ = (
    "Complete Framework for AI Chatbots and Agents, "
    "Supporting A2A, MCP, RAG and more."
)
__version__ = "0.24.14"
__author__ = "Jesus Lara"
__author_email__ = "jesuslarag@gmail.com"
__license__ = "MIT"
__copyright__ = "Copyright (c) 2020-2024 Jesus Lara"
