# parrot/exceptions.pyi
from typing import Any, Optional


class ParrotError(Exception):
    message: str
    stacktrace: Optional[Any]

    def __init__(self, message: Any, *args: Any, **kwargs: Any) -> None: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    def get(self) -> str: ...


class ConfigError(ParrotError): ...
class SpeechGenerationError(ParrotError): ...
class DriverError(ParrotError): ...
class ToolError(ParrotError): ...
